SystemOrganization addCategory: #'SimplePersistence-Core'!
SystemOrganization addCategory: #'SimplePersistence-Core-Developer'!

!Class methodsFor: '*SimplePersistence-Core-Developer' stamp: 'SeanDeNigris 2/4/2013 10:14'!
persist: aSymbol
	"
	aSymbol - name of the instanceVariable to persist
	See Object>>#persistIn:as: for more information"

	self persister persist: aSymbol in: self.! !

!Class methodsFor: '*SimplePersistence-Core-Developer' stamp: 'SeanDeNigris 2/1/2013 15:26'!
persister

	^ SpFileDatabase.! !

!Object methodsFor: '*SimplePersistence-Core-Developer' stamp: 'SeanDeNigris 2/1/2013 17:49'!
persistIn: aClass as: aSymbol
	"This is the main entry point to set up for persistence. It creates the database class for the package with all necessary methods, and the required methods in the domain class that holds the persistence object.
	aClass - will hold this object in a class inst var
	aSymbol - the name of that inst var
	"

	aClass persist: aSymbol.! !

Object subclass: #SpFileDatabase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core'!

SpFileDatabase class
	instanceVariableNames: 'lock'!

!SpFileDatabase commentStamp: 'SeanDeNigris 5/29/2011 13:36' prior: 0!
Based on the idea that:
* most applications will not have to scale (i.e. become the next Twitter)
* simply saving the image is slow and error-prone

For the full motivation, see Ramon Leon's blog post at http://onsmalltalk.com/simple-image-based-persistence-in-squeak/.

To give your application persistence:
	1. Subclass SMFileDatabase
	2. On the subclass, implement:
		a. class>>repositories (see method comment).
		b. class>>restoreRepositories: (see method comment).
		
That's it!! Now, whenever you want to save, call class>>saveRepository or class>>takeSnapshot (background save).

To customize:
* Number of backups kept: override class>>defaultHistoryCount
			!

SpFileDatabase class
	instanceVariableNames: 'lock'!

!SpFileDatabase class methodsFor: 'private' stamp: 'tbn 2/28/2013 12:10'!
availableVersions

	^ self backupDirectory files collect: [ :e | 
		self versionOf: e ].! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'tbn 3/1/2013 23:09'!
backupDirectory

	^ (FileLocator imageDirectory / self backupDirectoryName) ensureDirectory.
! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'tbn 3/1/2013 23:09'!
backupDirectoryName

	^self name! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 4/11/2012 23:05'!
backupFileForVersion: aNumber

	^ self backupDirectory files detect: [:each | (self versionOf: each) = aNumber ] ifNone: [ nil ].! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 2/4/2013 12:48'!
backupFiles

	^ self backupDirectory files select: [ :e | 
			(e basename beginsWith: self name) and: [ e extension isAllDigits ] ].! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 2/1/2013 15:24'!
backupVersion: aNumber
	| backup |
	backup := self backupFileForVersion: aNumber.
	backup ifNil: [ ^ nil ].
	^ self serializer materializeFrom: backup.! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 5/29/2011 13:22'!
defaultHistoryCount
	"How many backups to keep."
	^ 50! !

!SpFileDatabase class methodsFor: 'private' stamp: 'rjl 12/4/2007 18:15'!
enablePersistence
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 2/4/2013 12:47'!
lastBackupVersion

	| lastBackupFile |
	lastBackupFile := self backupFiles detectMax: [:each | self versionOf: each].
	^ lastBackupFile isNil
		ifTrue: [ 0 ]
		ifFalse: [ self versionOf: lastBackupFile ].
	! !

!SpFileDatabase class methodsFor: 'startup/shutdown' stamp: 'SeanDeNigris 2/1/2013 17:49'!
persist: aSymbol in: aClass 
	"See Object>>#persistIn:as:"

	SpFileDatabaseBehavior createFor: aClass persisting: aSymbol.
	SpPersistentBehavior addTo: aClass for: aSymbol.! !

!SpFileDatabase class methodsFor: 'restoring' stamp: 'SeanDeNigris 5/29/2011 13:27'!
repositories
	"Return the root object that needs serialized (e.g. an array containing the root collection of each domain class)."
	self subclassResponsibility! !

!SpFileDatabase class methodsFor: 'saving' stamp: 'SeanDeNigris 1/11/2012 19:55'!
restoreBackupVersion: aNumber
	(self backupVersion: aNumber) ifNotNilDo: [ :backup | self restoreRepositories: backup ]! !

!SpFileDatabase class methodsFor: 'saving' stamp: 'SeanDeNigris 1/11/2012 20:33'!
restoreLastBackup

	self restoreBackupVersion: self lastBackupVersion.! !

!SpFileDatabase class methodsFor: 'restoring' stamp: 'SeanDeNigris 5/29/2011 13:29'!
restoreRepositories: someRepositories
	"Put them back  wherever they belong in the image for the application to use them."
	self subclassResponsibility! !

!SpFileDatabase class methodsFor: 'setting up' stamp: 'SeanDeNigris 2/1/2013 14:10'!
saveRepository
	| version |
	lock ifNil: [ lock := Semaphore forMutualExclusion ].
	lock critical: [
		self trimBackups.
		version := self lastBackupVersion + 1.
		self serializer serialize: self repositories to: self backupDirectory / self name, version asString ].! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 2/1/2013 14:19'!
serializer
	"See SpSerializer class comment"

	^ SpFuelSerializer new.! !

!SpFileDatabase class methodsFor: 'private' stamp: 'rjl 12/4/2007 18:19'!
shutDown: isGoingDown 
	isGoingDown ifTrue: [ self saveRepository ]! !

!SpFileDatabase class methodsFor: 'private' stamp: 'rjl 12/4/2007 18:24'!
startUp: isComingUp 
	isComingUp ifTrue: [ self restoreLastBackup ]! !

!SpFileDatabase class methodsFor: 'saving' stamp: 'SeanDeNigris 5/29/2011 13:26'!
takeSnapshot
	"Motivation (vs. #restoreLastBackup): if you want a little extra snappiness and you're not worried about making the user wait for the flush to disk"
	[self saveRepository] forkAt: Processor userBackgroundPriority
		named: 'snapshot: ' , self class name! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 4/11/2012 23:15'!
trimBackups
	| entries versionsToKeep |
	versionsToKeep := self defaultHistoryCount.
	entries := self backupDirectory files select: [ :e | e base = self name ].
	entries size < versionsToKeep ifTrue: [ ^ self ].
	((entries sortBy: [ :a :b | a extension asInteger < b extension asInteger ]) allButLast: versionsToKeep) 
		do: [ :entry | entry delete ].! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 2/1/2013 15:43'!
versionOf: backupFile
	^ backupFile extension asInteger.! !

Object subclass: #SpFileDatabaseBehavior
	instanceVariableNames: 'containerClass instVarName databaseClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core-Developer'!

!SpFileDatabaseBehavior class methodsFor: 'setting up' stamp: 'SeanDeNigris 2/4/2013 10:21'!
createFor: aClass persisting: aSymbol 

	| instance |
	instance := self new
			containerClass: aClass;
			instVarName: aSymbol.
	^ instance create.! !

!SpFileDatabaseBehavior methodsFor: 'setting up' stamp: 'SeanDeNigris 2/4/2013 10:31'!
addInitializer
	"Call on subclasses to add class initialization"

	self databaseClass class compile:
'initialize

	self restoreLastBackup.
	self enablePersistence' classified: 'class initialization'.! !

!SpFileDatabaseBehavior methodsFor: 'setting up' stamp: 'SeanDeNigris 2/4/2013 10:34'!
addRepositoryAccessorFor: aSymbol of: aClass 

	| template source |
	template := 
'repositories

	^ \{ {1} {2} \}.'.
	source := template format: { aClass name. aSymbol }.
	self databaseClass class compile: source.! !

!SpFileDatabaseBehavior methodsFor: 'setting up' stamp: 'SeanDeNigris 2/4/2013 10:34'!
addRestorationMethodFor: aClass 

	| template source |
	template := 
'restoreRepositories: someRepositories

	{1} restoreFrom: someRepositories first.'.
	source := template format: { aClass name }.
	self databaseClass class compile: source.! !

!SpFileDatabaseBehavior methodsFor: 'private' stamp: 'SeanDeNigris 2/4/2013 10:25'!
containerClass: aClass

	containerClass := aClass.! !

!SpFileDatabaseBehavior methodsFor: 'as yet unclassified' stamp: 'SeanDeNigris 2/4/2013 10:32'!
create

	self
		addInitializer;
		addRepositoryAccessorFor: instVarName of: containerClass;
		addRestorationMethodFor: containerClass.! !

!SpFileDatabaseBehavior methodsFor: 'setting up' stamp: 'SeanDeNigris 2/4/2013 10:33'!
databaseClass
	
	| topLevelCategory dbClassName |
	databaseClass ifNil: [
		topLevelCategory := containerClass category copyUpTo: $-.
		dbClassName := (topLevelCategory, 'DB') asSymbol.
		databaseClass := SpFileDatabase subclass: dbClassName
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''
		category: topLevelCategory, '-Persistence' ].
	
	^ databaseClass.! !

!SpFileDatabaseBehavior methodsFor: 'private' stamp: 'SeanDeNigris 2/4/2013 10:25'!
instVarName: aSymbol

	instVarName := aSymbol.! !

Object subclass: #SpPersistentBehavior
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core-Developer'!

SpPersistentBehavior class
	instanceVariableNames: 'persistedClass persistedInstance'!

SpPersistentBehavior class
	instanceVariableNames: 'persistedClass persistedInstance'!

!SpPersistentBehavior class methodsFor: 'as yet unclassified' stamp: 'SeanDeNigris 12/8/2011 00:39'!
addTo: aClass for: aSymbol 

	^ self new addTo: aClass for: aSymbol.! !

!SpPersistentBehavior methodsFor: 'as yet unclassified' stamp: 'SeanDeNigris 2/4/2013 11:06'!
addTo: aClass for: aSymbol 

	| template source |
	template := 
'restoreFrom: anObject

	{1} := anObject.'.
	source := template format: { aSymbol }.
	aClass class compile: source classified: 'persistence'.! !

Object subclass: #SpSerializer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core'!

!SpSerializer commentStamp: 'SeanDeNigris 2/1/2013 14:19' prior: 0!
I am used by SpFileDatabase to do the actual serialization and materialization. Currently, Fuel is the only supported serializer. The original ReferenceStream serializer has been packaged as SimplePersistence-ReferenceStream just in case a legacy system requires it, but there is no hook to install it, so you'd have to override SpFileDatabase>>serializer.!

SpSerializer subclass: #SpFuelSerializer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core'!

!SpFuelSerializer commentStamp: 'SeanDeNigris 2/1/2013 14:20' prior: 0!
Uses Fuel: An open-source* general-purpose object serialization framework developed in Pharo Smalltalk environment.

See http://rmod.lille.inria.fr/web/pier/software/Fuel for more details.

* Developed under the MIT license.!

!SpFuelSerializer methodsFor: 'protocol' stamp: 'SeanDeNigris 2/1/2013 14:01'!
materializeFrom: aFileReference

	^ FLMaterializer materializeFromFileNamed: aFileReference fullName.! !

!SpFuelSerializer methodsFor: 'protocol' stamp: 'SeanDeNigris 2/1/2013 14:02'!
serialize: anObject to: aFileReference

	FLSerializer serialize: anObject toFileNamed: aFileReference fullName.! !

!SpSerializer methodsFor: 'protocol' stamp: 'SeanDeNigris 2/1/2013 13:56'!
materializeFrom: aFileReference

	self subclassResponsibility.! !

!SpSerializer methodsFor: 'protocol' stamp: 'SeanDeNigris 2/1/2013 14:02'!
serialize: anObject to: aFileReference

	self subclassResponsibility.! !
